
from markers import Markers
from options import getDirs,glbOptions

from sipylibs import FiltFiltFilter
from sipylibs.nestedgpint import NestedGPInterpolator
from sipylibs.qualisys import QualisysRawReader
from sipylibs import getPhaseEst
from sipylibs import PrintLogger
from sipylibs.utils import getRegions, getNeighbouringPoints

from pylab import figure,scatter,plot,savefig,title,xlabel,ylabel,close

from scipy.stats import nanmean
from numpy import isnan,logical_and,sum,mean,unwrap,pi,vstack,array
from numpy.random import seed

from os.path import isfile
from os.path import join as osjoin
from os import listdir
from yaml import load
from sys import argv

if len(argv)>=2 and argv[-1]!="--pylab":
  glb = glbOptions(dog=argv[-1])
else:
  print "WARNING: No dog specified, using default!"
  glb = glbOptions(None)


seed(0) # Ensure consistent results
dataDir,baseDir,outDir,intDir,logDir,plotDir = getDirs(glb.dog,glb.dataDir)

# Load the list of files we will be excluding

excludeList = [k['file'] for k in load(open(osjoin(glb.basePath,"excludelist.yaml"),'r'))]

# Get files to process
files = [f for f in listdir(dataDir)
                if f[-4:]==".tsv"
                and
                f[-7:]!="_an.tsv"
                and
                f not in excludeList]

for fn in files:
  zeronan = False
  ofle = osjoin(intDir,fn[:-4]+"_int.tsv")
  logfle = osjoin(logDir,fn[:-4]+"_int.log")
  # If already done, don't do again
  if isfile(ofle):
    continue
  
  # Create logger
  logger = PrintLogger(logfle)
  logger("Processing file: "+fn)
  
  # Load data
  fle = osjoin(dataDir,fn)
  Q = QualisysRawReader(fle,zeronan=zeronan)
  sr = Q.FREQUENCY

  # Create filters
  gentlelpf = FiltFiltFilter(sr,glb.highFreq,3,'lowpass')

  # Check NaN to zero settings, print warning
  # if ambiguous
  totNaNs = sum([sum(isnan(Q[k])) for k in Q])
  totZeros = sum([sum(logical_and(Q[k]<glb.eps,Q[k]>-glb.eps)) for k in Q])
  logger("Found "+str(totNaNs)+" NaNs and "+str(totZeros)+" zeros.")
  if totZeros>0 and zeronan==False:
    logger("WARNING: Zeros detected and not converted to NaNs!")    
    logger("Use NaN to zero (y/n)")
    if totNaNs==0: 
      zeronan = True
      logger("Converting zeros to Nans.")
      Q = QualisysRawReader(fle,zeronan=zeronan)
    else:
      logger("NaN also detected. Not converting zeros to Nans.")

  # Create a marker object for sorting out 
  # marker names
  m = Markers(Q.MARKER_NAMES)

  # Calculate leg phases for interpolation
  phi = dict()
  rx = dict()
  for k in m.appendages:
    x = Q[m.appendages[k]]
    x = x.reshape(x.shape[0],x.shape[1]*x.shape[2])
    x -= nanmean(x,0)
    phi[k] = getPhaseEst(x,verbose=False,fullOutput=False,minFreq=0.6,sr=sr)

  # Use average as body phase, but divide by four to ensure 
  # it has the correct frequency
  phi["Body"] = (mean([unwrap(phi[k]) for k in phi],0)/4)%(2*pi)

  # Make a list of things to fill
  toFill = dict(m.appendages)
  toFill["Body"] = m.bodyMarkers
  for k in toFill: # For each set of markers
    for q in toFill[k]: # For each marker
      x = Q[q]
      # Do we have any nans?
      if(sum(isnan(x))>0):
        logger(str(sum(isnan(x))/3)+" nans found in "+q+". Filling")
        figure()
        for ix,c in zip(x.T,['r','g','b']): # For every co-ordinate
          # Get regions that are nan and points 
          # that are nearby
          slicelist = getRegions(ix) 
          ptslist=getNeighbouringPoints(ix,slicelist,N=glb.NgPPts)
          # For every region with nans
          for slicetofill,valpts in zip(slicelist,ptslist):
            # X0 is the time+phase for valid neighboring points
            ptsIdx = array(valpts[:,0],dtype=int)
            X0=vstack([Q.Time[ptsIdx],phi[k][ptsIdx]]).T
            # Y0 is data for this valid region, or s[valptsidx]
            Y0 = valpts[:,1].reshape(valpts.shape[0],1)
            # Make an interpolator
            ngpi=NestedGPInterpolator(
                                      X0,
                                      Y0,
                                      glb.M,
                                      glb.sigmaSlow,
                                      glb.sigmaFast,
                                      glb.snoise,
                                      glb.sfac
                                      )
            # We should not "recursively" fill, this will compound interpolation errors.
            # Fill the points in the new signals.
            # Xstar is missing regions
            Xstar = vstack([Q.Time[slicetofill],phi[k][slicetofill]]).T
            # Fill
            ix[slicetofill]=ngpi.pred(Xstar)[:,0]
          # Now gently low pass filter to deal with 
          # the gp interpolator not being as strict 
          # about endpoints as we might like
          # Make a plot of the finished product
          filtx = gentlelpf(ix)
          for slicetofill in slicelist:
            ix[slicetofill] = filtx[slicetofill]
          for slicetofill in slicelist:
            scatter(Q.Time[slicetofill],ix[slicetofill],color='b',marker='x',s=9,lw=0.3)
          plot(Q.Time,ix,color=c)
        title("Marker: "+q)
        xlabel("time (s)")
        ylabel("position (mm)")
        graphfn = osjoin(plotDir,fn[:-4]+q+"_int.svg")
        logger("Writing graph to: "+graphfn)
        savefig(graphfn)
        if not glb.makePlots:
          close("all")
          
  # Save
  logger("Saving interpolated data at: "+ofle)
  Q.write(ofle)
