from sipylibs import zscoreConstants,FiltFiltFilter,zscore
from markers import Markers

from scipy.interpolate import UnivariateSpline
try:
  from scipy.stats import nanmean
except:
  from numpy import nanmean
from scipy.signal import hilbert
from numpy import (array,arange,std,hstack,sqrt,logical_not,dot,mean,median,
                   percentile,nan,isnan,sum,vstack,angle,where,logical_and,
                   cumsum,abs,ndarray)
from numpy.linalg import svd

def tukeyBiweight(x,c=3.0):
  D = (1-(x*x)/(c*c))
  if isinstance(x,ndarray):
    return(array(abs(x)<c,dtype=float)*x*D*D)
  else:
    return(float(abs(x)<c)*x*D*D)

def dtukeyBiweight(x,c=3.0):
  D = (1-(x*x)/(c*c))
  E = (1-5*(x*x)/(c*c))
  if isinstance(x,ndarray):
    return(array(abs(x)<c,dtype=float)*D*E)
  else:
    return(float(abs(x)<c)*D*E)

def splineDiff(x,dt=1.0):
  dx = list()
  for ix in x.T:
    t = arange(ix.size)
    s = UnivariateSpline(t,ix,s=0)
    dx.append(s(t+0.5)-s(t-0.5))
  return(array(dx).T/dt)

def gtProcess(x,com,hpf,lpf,sr):
  x = (x.transpose(1,0,2)-com).transpose(1,0,2)
  # Do PCA, the raw data actually gives the 
  # traces a reasonable weighting (things that 
  # swing big distances get more weight)
  x = x.reshape(x.shape[0],x.shape[1]*x.shape[2])
  # Gentle high pass filter to remove the 
  # offset for PCA
  x = hpf(x.T).T
  v = splineDiff(lpf(x.T).T,1.0/sr)
  # Weight velocity and position equally
  v = v/std(v)*std(x)
  v = hpf(v.T).T
  return(x,hstack([x,v]))

def getBeltSpeed(Q,floorCond,vlim=25.0):
  axis0 = [1,0,0]
  m = Markers(Q)
  vel = list()
  for leg in m.appendages:
    toeMarker = m.appendages[leg][0]
    toe = Q[toeMarker]
    vtoe = splineDiff(toe)
    nvtoe = (vtoe.T/sqrt(sum(vtoe**2,1))).T
    talgn = dot(axis0,nvtoe.T)
    fcFac = 1.0
    if mean(talgn[toe[:,2]<percentile(toe[:,2],50)])<0:
      fcFac = -1.0
    on_floor = ((toe[:,2]<percentile(toe[:,2],50))*talgn)*fcFac>floorCond
    vfoot = dot(axis0,vtoe.T)*on_floor
    vfoot[logical_not(on_floor)] = nan
    vfoot[abs(vfoot)>vlim] = nan
    vel.append(vfoot)
  vbelt = nanmean(vel,0)
  vbelt[abs(vbelt)>vlim] = nan
  # Prevent spline going haywire
  if isnan(vbelt[0]):
    idx = where(logical_not(isnan(vbelt)))[0][0]
    vbelt[:idx] = vbelt[idx]
  if isnan(vbelt[-1]):
    idx = where(logical_not(isnan(vbelt)))[0][-1]
    vbelt[idx:] = vbelt[idx]  
  tau = arange(vbelt.size)
  goodIdx = logical_not(isnan(vbelt))
  s = UnivariateSpline(tau[goodIdx],vbelt[goodIdx],s=0)#s=sum(goodIdx)*1*nanstd(vbelt))
  return(s(tau))

class Struct(object):
  pass

class GTTrans(object):
  '''
  After training this takes a qualisys object and spits out signals for phase 
  estimation for a given leg.
  '''
  def __init__(self,NPC,subset,lowFreq,highFreq,cut=0):
    self.mu0 = None
    self.sigma0 = None
    self.sigma1 = None # FIXME
    self.c = None # FIXME
    self.NPC = NPC # Number of principal components
    self.subset=subset
    self.lowFreq = lowFreq
    self.highFreq = highFreq
    self.cut = cut
      
  def train(self,Qs):
    ms = [Markers(Q) for Q in Qs]
    hpfs = [FiltFiltFilter(Q.FREQUENCY,self.lowFreq,4,'highpass') for Q in Qs]
    
    # Calculate constants
    self.mu0 = dict()
    self.sigma0 = dict()
    self.sigma1 = dict()
    self.muf = dict()
    self.sigmaf = dict()
    self.c = dict()
    coms = [mean(Q[m.bodyMarkers],1) for Q,m in zip(Qs,ms)]
    for k in ms[0].appendages:
      # Z-score constants
      xcoms = [
        (Q[m.appendages[k]].transpose(1,0,2)-com).transpose(1,0,2) 
        for com,Q,m in zip(coms,Qs,ms)
      ]
      # Reshape and filter
      xcoms = [
        hpf(xcom.reshape(xcom.shape[0],xcom.shape[1]*xcom.shape[2]).T).T
        for xcom,hpf in zip(xcoms,hpfs)
      ]
      self.mu0[k],self.sigma0[k] = zscoreConstants(vstack(xcoms).T)
      
      # Integral rescaling constant
      icoms = [self.oscIntegrate(xcom-self.mu0[k]) for xcom in xcoms]
      self.sigma1[k] = std(vstack(icoms),0)
      
      # PCA
      sig = vstack([
        hstack([
          x-self.mu0[k],
          v*(self.sigma0[k]/self.sigma1[k])
        ])
        for x,v in zip(xcoms,icoms)
      ])[:,self.subset]
      
      a,b,self.c[k] = svd(sig,False)
      
      # Now get median zscore constants
      r = dot(sig,self.c[k].T)[:,:self.NPC]
      self.muf[k], self.sigmaf[k] = median(r,0),std(r,0)
  
  def oscIntegrate(self,xcom):
    # Calculate integration constants
    a = angle(hilbert(zscore(array([xcom[:,0]]))[0]))
    i,j = where(logical_and(a[1:]>0,a[:-1]<0))[0][[0,-1]]
    # Integrate and scale
    icom = cumsum(xcom,0)
    return(icom-mean(icom[i:j],0))

  def getvCoM(self,Q):
    com = self.getCoM(Q)
    lpf = FiltFiltFilter(Q.FREQUENCY,self.highFreq,4,'lowpass')
    com = lpf(com.T).T
    return(splineDiff(com))
    
  
  def getCoM(self,Q):
    m = Markers(Q)
    x = Q[m.bodyMarkers]
    com = mean(x,1)
    return(com)    
  
  def getXCoM(self,Q):
    m = Markers(Q)
    hpf = FiltFiltFilter(Q.FREQUENCY,self.lowFreq,4,'highpass')
    lpf = FiltFiltFilter(Q.FREQUENCY,self.highFreq,4,'lowpass')
    # Calculate CoM
    com = self.getCoM(Q)
    xcom = dict()
    for k in m.appendages:
      x = Q[m.appendages[k]]    
      # Calculate CoM subtracted positions and integrate
      xcom[k] = (x.transpose(1,0,2)-com).transpose(1,0,2)
      # Reshape and filter
      xcom[k] = lpf(hpf(
        xcom[k].reshape(xcom[k].shape[0],xcom[k].shape[1]*xcom[k].shape[2]).T
      )).T
    return(xcom)
      
  def __call__(self,Q):
    m = Markers(Q)
    
    # Get CoM
    xcom = self.getXCoM(Q)
    
    # Grab leg data
    sig = dict()
    for k in m.appendages:
      icom = self.oscIntegrate(xcom[k]-self.mu0[k])
      # Center and stack and scale velocities to give right weight in PCA
      sig[k] = hstack([
        xcom[k]-self.mu0[k],
        icom*(self.sigma0[k]/self.sigma1[k])
      ])[:,self.subset]
    
      # Do PCA and zscore
      sig[k] = (dot(sig[k],self.c[k].T)[self.cut:(-self.cut if self.cut>0 else None),:self.NPC]-self.muf[k])/self.sigmaf[k]
    
    return(sig)
